/*******************************************************************************
 *
 *        Program : prs.c
 *
 *        Author  : Kim, Dongmin (kimd@top.cis.syr.edu)
 *
 * This program simulates the vector prefix sum and redunction.
 *
 * The vector prefix sum and reduction is like pure prefix sum and reduction
 *
 * except the exchanging data is vector not one variable.
 *
 * In this program, the prefix sum and reduction are combined in one routine
 *
 * so that the communication time is reduced, and
 *
 * two different algorithms are simulated.
 *
 * One is Naive algorithm mentioned in Kuma's Parallel Computing, and
 *
 * the other is Split algorithm.
 *
 * The main difference of these two algorithms is the step and
 *
 * the exchange data size.
 *
 * The Naive has lg N steps and exchanges N size of data.
 *
 * The Split has 2 * lg N steps and exchanges data starting N/2 and
 *
 * reduces the size by half, where N is the vector size.
 *
 * We simulates these two algorithms according to processor number and
 *
 * vector size, and see what happens.
 *
 * Some important procedures are like follows :
 *
 * prs_dim_naive() : Combined prefix and reduction in the Naive algorithm
 *
 * prs_dim_split() : Combined prefix and reduction in the Split algorithm
 *
 *******************************************************************************/

#include <cm/cmmd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "global.h"
#include "function.h"

/*******************************************************************************
 *   Procedure : read_input                                                    *
 *   Input     : None
 *   Output    : None
 *   Function  : read the input variable from the consol.
 *******************************************************************************/
read_input() {
   int error;
   int i, j;
   int total = 1, t_total=1;
   int save;

   ori_nproc = CMMD_partition_size();

   /*---------------- Rank of the input Array ---------------------------*/
   do {
      printf("\nEnter rank of the input array (Max=%d): ", Rankmax);
      scanf("%d", &rank);
      error = ( rank < 1 || rank > Rankmax );
      if (error)
	 printf("\n??? Improper rank. Try again!\n");
    } while (error);

   printf("\nEnter direction to communicate (0 ~ %d): ", rank-1);
   scanf("%d", &direction);

   printf("\nEnter vector size : ");
   scanf("%d", &v_size);

   /*-------------------- from dim=0 to dim=rank-1 ----------------------*/
   for (i=0; i<rank; ++i) {

     /*------------ Source Distribution for A, M ------------------*/
     save = total;
     do {
       printf("\n   Enter number of source processors (Max=%d): ", 
	      Procmax/total);
       scanf("%d", &nproc[i]);
       total *= nproc[i];
       error = ( nproc[i] < 1 || total > ori_nproc );
       dproc[i] = Ilog2(nproc[i]);
       if (error) {
	 total = save;
	 printf("\n??? Improper source processor number. Try again!\n");
       }
     } while (error);
   }

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   tot_nproc = total;
   tot_dproc = Ilog2(tot_nproc);

   /*======= Who I am ? =======*/
   my_pnum  = CMMD_self_address(); 
   map_to_grid(my_pnum, my_grid);

   /* setup the partner, right child and the parent */
   for ( i = 0 ; i < rank ; i++ ) {
     partner_grid[i] = my_grid[i];
     right_grid[i] = my_grid[i];
     parent_grid[i] = my_grid[i];
   }

   i = CMMD_reset_partition_size(tot_nproc);
   if (i != tot_nproc) 
     CMMD_error("\nERROR: CMMD Partition failed!\n");

   /* every processor use a different random seed */
   /*srandom(CMMD_self_address()); */

   if ( CMAML_allocate_this_rport(rport) <= 0 )
     CMMD_error("ERROR: Failed to allocate receive rport %d\n", rport);

}

/*******************************************************************************
 *   Procedure : PS_dim_naive()
 *   Input     : int dim, int size, int* p_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using naive algorithm        
 *               which is defined Kurmar's book
 *******************************************************************************/
int PS_dim_naive(dim, size, input_vector, p_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *p_result;    /* exclusive prefix */
{

  int partner, partner_grid[Rankmax];
  int i, j;
  int r_result[Max];

  for (j=0; j<size; ++j) {
    r_result[j] = input_vector[j];
    p_result[j] = 0;
  }

  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  CMAML_set_rport_total_nbytes (rport, msg_size*size);

  CMMD_sync_with_nodes();

  for (j=0; j<rank; ++j)
    partner_grid[j] = my_grid[j];

  for (i=0; i<dproc[dim]; ++i) {
  
    partner_grid[dim] = my_grid[dim] ^ pow2(i);
    map_to_pnum(partner_grid, &partner);

    prs_dim_ack = false;

    CMAML_set_rport_byte_counter (rport, msg_size*size);

    CMMD_sync_with_nodes();

    CMAML_scopy(partner,
		rport,
		0,
		0,
		r_result,
		msg_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    for (j=0; j<size; ++j)
      r_result[j] += in_mesg[j];

    if (partner_grid[dim] < my_grid[dim])
      for (j=0; j<size; ++j)
	p_result[j] += in_mesg[j];

    partner_grid[dim] = my_grid[dim];

  }

  CMAML_enable_interrupts();

/****
  show(my_pnum, 0, size, p_result);
****/
}

/*******************************************************************************
 *   Procedure : RS_dim_naive()
 *   Input     : int dim, int size, int* r_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using naive algorithm        
 *               which is defined Kurmar's book
 *******************************************************************************/
int RS_dim_naive(dim, size, input_vector, r_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *r_result;    /* reduction */
{

  int partner, partner_grid[Rankmax];
  int i, j;

  for (j=0; j<size; ++j) {
    r_result[j] = input_vector[j];
  }

  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  CMAML_set_rport_total_nbytes (rport, msg_size*size);

  CMMD_sync_with_nodes();

  for (j=0; j<rank; ++j)
    partner_grid[j] = my_grid[j];

  for (i=0; i<dproc[dim]; ++i) {
  
    partner_grid[dim] = my_grid[dim] ^ pow2(i);
    map_to_pnum(partner_grid, &partner);

    prs_dim_ack = false;

    CMAML_set_rport_byte_counter (rport, msg_size*size);

    CMMD_sync_with_nodes();

    CMAML_scopy(partner,
		rport,
		0,
		0,
		r_result,
		msg_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    for (j=0; j<size; ++j)
      r_result[j] += in_mesg[j];

    partner_grid[dim] = my_grid[dim];

  }

  CMAML_enable_interrupts();

/****
  show(my_pnum, 0, size, r_result);
****/
}

/*******************************************************************************
 *   Procedure : PRS_dim_naive()
 *   Input     : int dim, int size, int* p_result, int* r_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using naive algorithm        
 *               which is defined Kurmar's book
 *******************************************************************************/
int PRS_dim_naive(dim, size, input_vector, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{

  int partner, partner_grid[Rankmax];
  int i, j;

  for (j=0; j<size; ++j) {
    r_result[j] = input_vector[j];
    p_result[j] = 0;
  }

  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  CMAML_set_rport_total_nbytes (rport, msg_size*size);

  CMMD_sync_with_nodes();

  for (j=0; j<rank; ++j)
    partner_grid[j] = my_grid[j];

  for (i=0; i<dproc[dim]; ++i) {
  
    partner_grid[dim] = my_grid[dim] ^ pow2(i);
    map_to_pnum(partner_grid, &partner);

    prs_dim_ack = false;

    CMAML_set_rport_byte_counter (rport, msg_size*size);

    CMMD_sync_with_nodes();

    CMAML_scopy(partner,
		rport,
		0,
		0,
		r_result,
		msg_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    for (j=0; j<size; ++j)
      r_result[j] += in_mesg[j];

    if (partner_grid[dim] < my_grid[dim])
      for (j=0; j<size; ++j)
	p_result[j] += in_mesg[j];

    partner_grid[dim] = my_grid[dim];

  }

  CMAML_enable_interrupts();

/****
  show(my_pnum, 0, size, p_result);
  show(my_pnum, 0, size, r_result);
****/
}

/*******************************************************************************
 *   Procedure : PS_dim_split()
 *   Input     : int vector_size, int input_v, int* p_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using advanced algorithm
 *******************************************************************************/
int PS_dim_split(dim, vector_size, input_vector, p_result)
     int dim,          /* PRS along dimension = dim */
         vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *p_result;    /* prefix sum output array */
{

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    tmp_p[Max];

  /* setup the partner, right child and the parent */
  for ( i = 0 ; i < rank ; i++ ) {
    partner_grid[i] = my_grid[i];
  }

  /* setup the vectors */
  for ( i = 0 ; i < vector_size ; i++ ) {
    tmp_p[i] = input_vector[i];
  }

/***
  show(my_pnum, 0, vector_size, r_result);
***/

  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(nproc[dim]);

  /** For Active Message Passing **/
  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();
/* ===============
       SUM
   =============== */

  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;
    start_send = start_recv;
    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &tmp_p[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
	Left[step][i] = tmp_p[i];
	tmp_p[i] += in_mesg[j];
      }
    }
    else {
      for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
	tmp_p[i] += in_mesg[j];
	Left[step][i] = in_mesg[j];
      }
    }
  }

/****
  show(my_pnum, 0, vector_size, tmp_p);
  show(my_pnum, 0, vector_size, Right[0]);
****/
  for ( i = start_recv ; i < start_recv + data_size ; i++ ) {
    p_result[i] = 0;
  }

  start_send = start_recv;

/* ===========
     PRESFIX
   =========== */

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = start_send, j = 0 ; i < start_send + data_size ; i++, j++ ) {
	out_mesg[j] = p_result[i] + Left[step][i];
      }
    }
    else {
      for ( i = start_send, j = 0 ; i < start_send + data_size ; i++, j++ ) {
	out_mesg[j] = p_result[i];
	p_result[i] += Left[step][i];
      }
    }

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
/***
    CMAML_set_rport_addr         (rport, in_mesg);
***/
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		out_mesg,
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
      p_result[i] = in_mesg[j];
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

/****
  show(my_pnum, 0, vector_size, p_result);
****/

  return 0;

}

/*******************************************************************************
 *   Procedure : RS_dim_split()
 *   Input     : int vector_size, int input_v, int* r_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using advanced algorithm
 *******************************************************************************/
int RS_dim_split(dim, vector_size, input_vector, r_result)
     int dim,          /* PRS along dimension = dim */
         vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *r_result;    /* reduction output array */
{

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;

  /* setup the partner, right child and the parent */
  for ( i = 0 ; i < rank ; i++ ) {
    partner_grid[i] = my_grid[i];
  }

  /* setup the vectors */
  for ( i = 0 ; i < vector_size ; i++ ) {
    r_result[i] = input_vector[i];
  }

/***
  show(my_pnum, 0, vector_size, r_result);
***/

  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(nproc[dim]);

  /** For Active Message Passing **/
  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();
/* ===============
       SUM
   =============== */

  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;
    start_send = start_recv;
    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &r_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
      r_result[i] += in_mesg[j];
    }
  }

/****
  show(my_pnum, 0, vector_size, r_result);
  show(my_pnum, 0, vector_size, Right[0]);
****/

  start_send = start_recv;

/* =============
     REDUCTION
   ============= */

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
/***
    CMAML_set_rport_addr         (rport, in_mesg);
***/
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		&r_result[start_send],
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
      r_result[i] = in_mesg[j];
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

/****
  show(my_pnum, 0, vector_size, r_result);
****/

  return 0;

}

/*******************************************************************************
 *   Procedure : PRS_dim_split_normal()
 *   Input     : int vector_size, int input_v, int* p_result, int* r_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using advanced algorithm
 *******************************************************************************/
int PRS_dim_split_normal(dim, vector_size, input_vector, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *p_result,    /* prefix sum output array */
         *r_result;    /* reduction output array */
{

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    save_recv;
  int    save_data_size;

  /* setup the partner, right child and the parent */
  for ( i = 0 ; i < rank ; i++ ) {
    partner_grid[i] = my_grid[i];
  }

  /* setup the vectors */
  for ( i = 0 ; i < vector_size ; i++ ) {
    r_result[i] = input_vector[i];
  }

/***
  show(my_pnum, 0, vector_size, r_result);
***/

  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(nproc[dim]);

  /** For Active Message Passing **/
  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();
/* ===============
       SUM
   =============== */

  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;
    start_send = start_recv;
    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &r_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
	Left[step][i] = r_result[i];
	r_result[i] += in_mesg[j];
      }
    }
    else {
      for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
	r_result[i] += in_mesg[j];
	Left[step][i] = in_mesg[j];
      }
    }
  }

/****
  show(my_pnum, 0, vector_size, r_result);
  show(my_pnum, 0, vector_size, Right[0]);
****/
  for ( i = start_recv ; i < start_recv + data_size ; i++ ) {
    p_result[i] = 0;
  }

  save_recv = start_recv;
  save_data_size = data_size;
  start_send = start_recv;

/* ===========
     PRESFIX
   =========== */

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = start_send, j = 0 ; i < start_send + data_size ; i++, j++ ) {
	out_mesg[j] = p_result[i] + Left[step][i];
      }
    }
    else {
      for ( i = start_send, j = 0 ; i < start_send + data_size ; i++, j++ ) {
	out_mesg[j] = p_result[i];
	p_result[i] += Left[step][i];
      }
    }

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
/***
    CMAML_set_rport_addr         (rport, in_mesg);
***/
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		out_mesg,
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
      p_result[i] = in_mesg[j];
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

  start_recv = save_recv;
  data_size = save_data_size;
  start_send = start_recv;

/* =============
     REDUCTION
   ============= */

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
/***
    CMAML_set_rport_addr         (rport, in_mesg);
***/
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		&r_result[start_send],
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
      r_result[i] = in_mesg[j];
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

/****
  show(my_pnum, 0, vector_size, p_result);
  show(my_pnum, 0, vector_size, r_result);
****/

  return 0;

}

/*******************************************************************************
 *   Procedure : PRS_dim_split_op()
 *   Input     : int vector_size, int input_v, int* p_result, int* r_result
 *   Output    : int
 *   Function  : Scalable vector PRS on PS_i & RS_i using advanced algorithm
 *******************************************************************************/
int PRS_dim_split_op(dim, vector_size, input_vector, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *input_vector, /* input array */
         *p_result,    /* prefix sum output array */
         *r_result;    /* reduction output array */
{

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;

  if ( p_result == r_result ) {
    return -1;
  }

  /* setup the partner, right child and the parent */
  for ( i = 0 ; i < rank ; i++ ) {
    partner_grid[i] = my_grid[i];
  }

  /* setup the vectors */
  for ( i = 0 ; i < vector_size ; i++ ) {
    r_result[i] = input_vector[i];
  }

/***
  show(my_pnum, 0, vector_size, r_result);
***/

  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(nproc[dim]);

  /** For Active Message Passing **/
  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();
/* ===============
       SUM
   =============== */

  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;
    start_send = start_recv;
    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &r_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

/***
    for ( i = start_recv ; i < start_recv + data_size ; i++ ) {
      Right[step][i] = p_result[i];
      p_result[i] += in_mesg[i];
    }
***/

    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
	Left[step][i] = r_result[i];
	r_result[i] += in_mesg[j];
      }
    }
    else {
      for ( i = start_recv, j = 0 ; i < start_recv + data_size ; i++, j++ ) {
	r_result[i] += in_mesg[j];
	Left[step][i] = in_mesg[j];
      }
    }
  }

/****
    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = 0 ; i < data_size ; i++ ) {
	p_result[i + start_recv] += in_mesg[i];
	Right[step][i + start_recv] = in_mesg[i];
      }
    }
    else {
      for ( i = 0 ; i < data_size ; i++ ) {
	Right[step][i + start_recv] = p_result[i + start_recv];
	p_result[i + start_recv] += in_mesg[i];
      }
    }
  }
****/

/****
  show(my_pnum, 0, vector_size, r_result);
  show(my_pnum, 0, vector_size, Right[0]);
****/

  for ( i = start_recv ; i < start_recv + data_size ; i++ ) {
    p_result[i] = 0;
  }
  start_send = start_recv;

/* =======================
     PRESFIX & REDUCTION
   ======================= */

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_grid[dim] & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_grid[dim] = my_grid[dim] ^ pow2(step);
    map_to_pnum(partner_grid, &partner_pnum);

    if ( my_grid[dim] < partner_grid[dim] ) {
      for ( i = start_send, j = 0, k = data_size ; i < start_send + data_size ; i++, j++, k++ ) {
	out_mesg[j] = p_result[i] + Left[step][i];
	out_mesg[k] = r_result[i];
      }
    }
    else {
      for ( i = start_send, j = 0, k = data_size ; i < start_send + data_size ; i++, j++, k++ ) {
	out_mesg[j] = p_result[i];
	p_result[i] += Left[step][i];
	out_mesg[k] = r_result[i];
      }
    }

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
/***
    CMAML_set_rport_addr         (rport, in_mesg);
***/
    CMAML_set_rport_byte_counter (rport, data_size*size*2);
    CMAML_set_rport_total_nbytes (rport, data_size*size*2);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		out_mesg,
		data_size*size*2,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = start_recv, j = 0, k = data_size ; i < start_recv + data_size ; i++, j++, k++ ) {
      p_result[i] = in_mesg[j];
      r_result[i] = in_mesg[k];
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

/****
  show(my_pnum, 0, vector_size, p_result);
  show(my_pnum, 0, vector_size, r_result);
****/

  return 0;

}

/*******************************************************************************
 *                                                                             * 
 *                              MAIN  PROGRAM                                  *
 *                                                                             *
 *******************************************************************************/
  
main(argc, argv)
     int argc;
     char *argv[];
{
  FILE   *fd; /* input vector file name */

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  double time, time0, time1, time2;
  double time3, time4, time5, time6;
  int    p_result[Max], r_result[Max], input_vector[Max];
  int    input_v[Max];
  double cmmd, prs;

  /*-------------------------  Initialize  -----------------------------*/

  read_input();

  CMMD_node_timer_clear(0);
  CMMD_node_timer_clear(1);
  CMMD_node_timer_clear(2);
  CMMD_node_timer_clear(3);
  CMMD_node_timer_clear(4);
  CMMD_node_timer_clear(5);
  CMMD_node_timer_clear(6);

  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(0);
  PRS_dim_naive(direction, v_size, input_vector, p_result, r_result);
  CMMD_node_timer_stop(0);

  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(1);
  PS_dim_naive(direction, v_size, input_vector, p_result, r_result);
  CMMD_node_timer_stop(1);

  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(2);
  RS_dim_naive(direction, v_size, input_vector, p_result, r_result);
  CMMD_node_timer_stop(2);


  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(3);
  PS_dim_split(direction, v_size, input_vector, p_result);
  CMMD_node_timer_stop(3);

  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(4);
  RS_dim_split(direction, v_size, input_vector, r_result);
  CMMD_node_timer_stop(4);

  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(5);
  PRS_dim_split_normal(direction, v_size, input_vector, p_result, r_result);
  CMMD_node_timer_stop(5);

  for ( i = 0 ; i < v_size ; i++ ) {
    input_vector[i]  = 2;
  }
  CMMD_node_timer_start(6);
  PRS_dim_split_op(direction, v_size, input_vector, p_result, r_result);
  CMMD_node_timer_stop(6);

  time0 = CMMD_node_timer_busy(0) * 1000.;  /* milisec */
  time1 = CMMD_node_timer_busy(1) * 1000.;  /* milisec */
  time2 = CMMD_node_timer_busy(2) * 1000.;  /* milisec */
  time3 = CMMD_node_timer_busy(3) * 1000.;  /* milisec */
  time4 = CMMD_node_timer_busy(4) * 1000.;  /* milisec */
  time5 = CMMD_node_timer_busy(5) * 1000.;  /* milisec */
  time6 = CMMD_node_timer_busy(6) * 1000.;  /* milisec */

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n**************************************************************\n");
  printf("\n  Nproc       = %d", tot_nproc);
  printf("\n  Vector size = %d\n", v_size);
  printf("\n--------------------------------------------------------------\n");
  printf("\n  CMMD Naive PS  Algorithm           = %5.3f msec", time0);
  printf("\n  CMMD Naive RS  Algorithm           = %5.3f msec", time1);
  printf("\n  CMMD Naive PRS Algorithm           = %5.3f msec", time2);
  printf("\n--------------------------------------------------------------\n");
  printf("\n  CMMD Split PS  Algorithm           = %5.3f msec", time3);
  printf("\n  CMMD Split RS  Algorithm           = %5.3f msec", time4);
  printf("\n  CMMD Split PRS - normal  Algorithm = %5.3f msec", time5);
  printf("\n  CMMD Split PRS - optimal Algorithm = %5.3f msec", time6);
  printf("\n\n**************************************************************\n\n");

}
